#include <cassert>
#include "creature.h"
#include "random.h"

extern BITMAP * gcur;

int col_in [TYPE_NUM], col_ex [TYPE_NUM], col_dir [TYPE_NUM],
 delta [TYPE_NUM];

extern creature * crs [MaxN];
extern int board [BoardX] [BoardY];
extern int level;

void normalize (real & angle)
{
 while (angle > M2PI) angle -= M2PI;
 while (angle < 0.0) angle += M2PI;
}

bool inside (int x, int y)
{
 return (x >= 0 && x < BoardX && y >= 0 && y < BoardY);
}

void recur_effect_up (int x, int y)
{
 if (x > 0 &&          y > 0 &&          board[x - 1][y - 1] < board[x][y] - 3)
 {
  board[x - 1][y - 1]++;
  board[x][y]--;
  recur_effect_up (x - 1, y - 1);
 }
 if (x > 0 &&                            board[x - 1][y    ] < board[x][y] - 2)
 {
  board[x - 1][y    ]++;
  board[x][y]--;
  recur_effect_up (x - 1, y    );
 }
 if (x > 0 &&          y < BoardY - 1 && board[x - 1][y + 1] < board[x][y] - 3)
 {
  board[x - 1][y + 1]++;
  board[x][y]--;
  recur_effect_up (x - 1, y + 1);
 }
 if (                  y < BoardY - 1 && board[x    ][y + 1] < board[x][y] - 2)
 {
  board[x    ][y + 1]++;
  board[x][y]--;
  recur_effect_up (x    , y + 1);
 }
 if (x < BoardX - 1 && y < BoardY - 1 && board[x + 1][y + 1] < board[x][y] - 3)
 {
  board[x + 1][y + 1]++;
  board[x][y]--;
  recur_effect_up (x + 1, y + 1);
 }
 if (x < BoardX - 1 &&                   board[x + 1][y    ] < board[x][y] - 2)
 {
  board[x + 1][y    ]++;
  board[x][y]--;
  recur_effect_up (x + 1, y    );
 }
 if (x < BoardX - 1 && y > 0 &&          board[x + 1][y - 1] < board[x][y] - 3)
 {
  board[x + 1][y - 1]++;
  board[x][y]--;
  recur_effect_up (x + 1, y - 1);
 }
 if (                  y > 0 &&          board[x    ][y - 1] < board[x][y] - 2)
 {
  board[x    ][y - 1]++;
  board[x][y]--;
  recur_effect_up (x    , y - 1);
 }
}

void recur_effect_down (int x, int y)
{
 if (x > 0 &&          y > 0 &&          board[x - 1][y - 1] > board[x][y] + 3)
 {
  board[x - 1][y - 1]--;
  board[x][y]++;
  recur_effect_down (x - 1, y - 1);
 }
 if (x > 0 &&                            board[x - 1][y    ] > board[x][y] + 2)
 {
  board[x - 1][y    ]--;
  board[x][y]++;
  recur_effect_down (x - 1, y    );
 }
 if (x > 0 &&          y < BoardY - 1 && board[x - 1][y + 1] > board[x][y] + 3)
 {
  board[x - 1][y + 1]--;
  board[x][y]++;
  recur_effect_down (x - 1, y + 1);
 }
 if (                  y < BoardY - 1 && board[x    ][y + 1] > board[x][y] + 2)
 {
  board[x    ][y + 1]--;
  board[x][y]++;
  recur_effect_down (x    , y + 1);
 }
 if (x < BoardX - 1 && y < BoardY - 1 && board[x + 1][y + 1] > board[x][y] + 3)
 {
  board[x + 1][y + 1]--;
  board[x][y]++;
  recur_effect_down (x + 1, y + 1);
 }
 if (x < BoardX - 1 &&                   board[x + 1][y    ] > board[x][y] + 2)
 {
  board[x + 1][y    ]--;
  board[x][y]++;
  recur_effect_down (x + 1, y    );
 }
 if (x < BoardX - 1 && y > 0 &&          board[x + 1][y - 1] > board[x][y] + 3)
 {
  board[x + 1][y - 1]--;
  board[x][y]++;
  recur_effect_down (x + 1, y - 1);
 }
 if (                  y > 0 &&          board[x    ][y - 1] > board[x][y] + 2)
 {
  board[x    ][y - 1]--;
  board[x][y]++;
  recur_effect_down (x    , y - 1);
 }
}

creature::creature (real nx, real ny, real nv, real na, real ns,
 int ntype, int nai)
{
 x = nx;
 y = ny;
 v = nv;
 a = na;
 s = ns;
 type = ntype;
 ai = nai;
 z = 0;
 w = 0.0;
}

void creature::move (void)
{
 x += w * cos (a);
 y += w * sin (a);
 control_flags = 0;
}

bool creature::check (void)
{
 if (x < s) {x = 2.0 * s - x; a = M1PI - a;}
 if (x > BoardX - s) {x = 2.0 * (BoardX - s) - x; a = M1PI - a;}
 if (y < s) {y = 2.0 * s - y; a = -a;}
 if (y > BoardY - s) {y = 2.0 * (BoardY - s) - y; a = -a;}
 normalize (a);
 return true;
}

void creature::accelerate (void)
{
 if (!(control_flags & 1))
 {
  control_flags |= 1;
  v += VDelta;
  if (v > VMax)
  v = VMax;
 }
}

void creature::brake (void)
{
 if (!(control_flags & 1))
 {
  control_flags |= 1;
  v -= VDelta;
  if (v < VMin)
   v = VMin;
 }
}

void creature::turn_left (void)
{
 if (!(control_flags & 2))
 {
  control_flags |= 2;
  a -= ADelta;
 }
}

void creature::turn_right (void)
{
 if (!(control_flags & 2))
 {
  control_flags |= 2;
  a += ADelta;
 }
}

void creature::control (void)
{
 switch (ai) // braces are for independent variable declarations
 {
  case AI_NONE:
  break;

  case AI_RANDOM:
  {
   int k;
   k = rndvalue (5);
   if (k == 0) turn_left ();
   else if (k == 4) turn_right ();
   k = rndvalue (6);
   if (k <= 1) accelerate ();
   else if (k == 5) brake ();
  }
  break;

  case AI_COWARD:
  {
   if (rndvalue (256) == 0) ai = AI_NEUTRAL;
   const int LOOKAHEAD = 15;
   int i, x1, y1, m;
   m = MaxB;
   for (i = 1; i <= LOOKAHEAD; i++)
   {
    x1 = x + cos (a) * i;
    y1 = y + sin (a) * i;
    if (!inside (x1, y1)) break;
    if (m > board[x1][y1])
     m = board[x1][y1];
   }
   if (m < -MidB)
   {
    brake ();
    real b;
    if (hypot (crs[0] -> y - y, crs[0] -> x - x) < eps) b = 0.0;
    else b = atan2 (crs[0] -> y - y, crs[0] -> x - x);
    while (b < a) b += M2PI;
    if (b - a < M1PI) turn_left ();
    else turn_right ();
   }
   else if (m < MidB && hypot (crs[0] -> y - y, crs[0] -> x - x) < 30.0)
   {
    accelerate ();
    real b;
    if (hypot (crs[0] -> y - y, crs[0] -> x - x) < eps) b = 0.0;
    else b = atan2 (crs[0] -> y - y, crs[0] -> x - x);
    while (b < a) b += M2PI;
    if (b - a < M1PI) turn_left ();
    else turn_right ();
   }
   else
   {
    accelerate ();
    if (rndvalue (128) == 0) ai = AI_AGGRESSIVE;
   }
  }
  break;

  case AI_AGGRESSIVE:
  {
   if (rndvalue (256) == 0) ai = AI_NEUTRAL;
   const int LOOKAHEAD = 10;
   int i, x1, y1, m;
   m = MaxB;
   for (i = 1; i <= LOOKAHEAD; i++)
   {
    x1 = x + cos (a) * i;
    y1 = y + sin (a) * i;
    if (!inside (x1, y1)) break;
    if (m > board[x1][y1])
     m = board[x1][y1];
   }
   if (m < -MidB)
   {
    brake ();
    real b;
    if (hypot (crs[0] -> y - y, crs[0] -> x - x) < eps) b = 0.0;
    else b = atan2 (crs[0] -> y - y, crs[0] -> x - x);
    while (b < a) b += M2PI;
    if (b - a < M1PI) turn_left ();
    else turn_right ();
    if (rndvalue (64) == 0) ai = AI_COWARD;
   }
   else if (m < MidB && hypot (crs[0] -> y - y, crs[0] -> x - x) < 30.0)
   {
    accelerate ();
    real b;
    if (hypot (crs[0] -> y - y, crs[0] -> x - x) < eps) b = 0.0;
    else b = atan2 (crs[0] -> y - y, crs[0] -> x - x);
    while (b < a) b += M2PI;
    if (b - a < M1PI) turn_left ();
    else turn_right ();
   }
   else
   {
    accelerate ();
    real b;
    if (hypot (crs[0] -> y - y, crs[0] -> x - x) < eps) b = 0.0;
    else b = atan2 (crs[0] -> y - y, crs[0] -> x - x);
    while (b < a) b += M2PI;
    if (b - a < M1PI) turn_right ();
    else turn_left ();
   }
  }
  break;

  case AI_NEUTRAL:
  {
   if (rndvalue (64) == 0) ai = AI_COWARD;
   if (rndvalue (128) == 0) ai = AI_AGGRESSIVE;
   const int LOOKAHEAD = 10;
   int i, x1, y1, m;
   m = MaxB;
   for (i = 1; i <= LOOKAHEAD; i++)
   {
    x1 = x + cos (a) * i;
    y1 = y + sin (a) * i;
    if (!inside (x1, y1)) break;
    if (m > board[x1][y1])
     m = board[x1][y1];
   }
   if (m < -MidB)
   {
    brake ();
    real b;
    if (hypot (crs[0] -> y - y, crs[0] -> x - x) < eps) b = 0.0;
    else b = atan2 (crs[0] -> y - y, crs[0] -> x - x);
    while (b < a) b += M2PI;
    if (b - a < M1PI) turn_left ();
    else turn_right ();
    if (rndvalue (128) == 0) ai = AI_COWARD;
   }
   else if (m < MidB && hypot (crs[0] -> y - y, crs[0] -> x - x) < 30.0)
   {
    accelerate ();
    real b;
    if (hypot (crs[0] -> y - y, crs[0] -> x - x) < eps) b = 0.0;
    else b = atan2 (crs[0] -> y - y, crs[0] -> x - x);
    while (b < a) b += M2PI;
    if (b - a < M1PI) turn_left ();
    else turn_right ();
   }
   else
   {
    int k;
    k = rndvalue (5);
    if (k == 0) turn_left ();
    else if (k == 4) turn_right ();
    k = rndvalue (6);
    if (k <= 1) accelerate ();
    else if (k == 5) brake ();
   }
  }
  break;

  default:
  {
   assert (false);
  }
  break;
 }
}

void creature::effect (void)
{
 int i, j;
 const int x0 = x, y0 = y, s0 = floor (s);
 for (i = -s0; i <= s0; i++)
  for (j = -s0; j <= s0; j++)
   if (i * i + j * j <= s0 * s0)
     {
      int & t = board[x0 + i][y0 + j];
      t += delta[type];
      if (t < -MaxB) t = -MaxB;
      if (t >  MaxB) t =  MaxB;
      if (delta[type] > 0)
       recur_effect_up   (x0 + i, y0 + j);
      else
       recur_effect_down (x0 + i, y0 + j);
     }
}

bool creature::suffer (void)
{
 int i, j, c, t;
 c = t = 0;
 const int x0 = x, y0 = y, s0 = floor (s);
 for (i = -s0; i <= s0; i++)
  for (j = -s0; j <= s0; j++)
   if (i * i + j * j <= s0 * s0)
   {
    c++;
    t += board[x0 + i][y0 + j];
   }
 assert (c > 0);
 z = t / c;
 w = v;
 if (type == TYPE_MONSTER)
 {
  if (w > (double) (VMax * (MaxB - z)) / (MaxB << 1))
   w = (double) (VMax * (MaxB - z)) / (MaxB << 1);
 }
 else
 {
  if (w > (double) (VMax * (z + MaxB)) / (MaxB << 1))
   w = (double) (VMax * (z + MaxB)) / (MaxB << 1);
 }
 if (type == TYPE_MONSTER)
  return z >= MidB;
 else
  return z <= -MidB;
}

void creature::draw (void)
{
 circlefill (gcur, ShiftX + x, ShiftY + y, s, col_in[type]);
 line (gcur, ShiftX + x, ShiftY + y,
  ShiftX + x + s * cos (a), ShiftY + y + s * sin (a), col_dir[type]);
 circle (gcur, ShiftX + x, ShiftY + y, s, col_ex[type]);
}
