/* This program uses Allegro library version 4.1.x or higher.
   Visit http://allegro.cc for more information on the library. */
#include <cassert>
#include <cmath>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <allegro.h>
#include "creature.h"
#include "global.h"
#include "random.h"
#include "score.h"
#include "types.h"

int colour [COLOUR_NUM];

extern int col_in [TYPE_NUM], col_ex [TYPE_NUM], col_dir [TYPE_NUM],
 delta [TYPE_NUM];

extern int ch_score [MaxCh] [MaxScore];
extern int lv_score [MaxLv] [MaxScore];
extern char ch_sname [MaxCh] [MaxScore] [MaxInput];
extern char lv_sname [MaxLv] [MaxScore] [MaxInput];

const int NumStr = 4;

const char str_startb [NumStr] [MaxBuf] =
 {"Water is about to end now.",
  "Now we fight, you waterbag!",
  "Liquids are weak.",
  "Boulders are strong."};

const char str_startm [NumStr] [MaxBuf] =
 {"Now, drown!",
  "The Ocean shall consume you all.",
  "Water will win this day!",
  "I shall sink all the land."};

const char str_win    [NumStr] [MaxBuf] =
 {"(Gulp!)...",
  "That won't be so easy next time!",
  "A whole ocean to win... how lame.",
  "Hope you drown in there yourself."};

const char str_lose   [NumStr] [MaxBuf] =
 {"We shall now dry you, forever.",
  "The Earth shall ever prevail!",
  "Feel your body get dry...",
  "Your tiny lake dried up, huh?"};

char buf [MaxBuf];
char name [MaxInput], level_name [MaxInput];
bool win, lose, End1;

BITMAP * gbuf [2], * gcur;
FONT * curfont;
int gb_cur;
bool kb_poll;

creature * crs [MaxN];
int board [BoardX] [BoardY];
int level;

volatile int tick;
int moves, score_pos;

void timer (void)
{
 tick++;
}
END_OF_FUNCTION (timer);

inline void poll_kb (void)
{
 if (kb_poll) poll_keyboard ();
}

void try_init (bool res, const char * str)
{
 if (res) printf ("%s init OK.\n", str);
 else
 {
  printf ("%s init failed!\n", str);
  exit (EXIT_FAILURE);
 }
}

void init_kb (void)
{
 try_init (install_keyboard () != -1, "Keyboard");
 kb_poll = keyboard_needs_poll ();
}

void to_close (void)
{
 End1 = true;
}

void init (void)
{
 freopen ("dat/log.txt", "wt", stdout);

 try_init (allegro_init () == 0, "Allegro");
 init_kb ();
 set_color_depth (32);
 try_init (set_gfx_mode (GFX_AUTODETECT_WINDOWED, MaxX, MaxY, 0, 0) == 0,
  "Graphics");
 try_init ((curfont = load_bitmap_font ("dat/8514fix.pcx", NULL, NULL))
  != NULL, "Font");
 set_window_title ("Earth and Water");
 set_close_button_callback (to_close);
 set_display_switch_mode (SWITCH_BACKGROUND);
 load_score ();

 try_init (install_timer () == 0, "Timer");
 LOCK_VARIABLE (tick);
 LOCK_FUNCTION (timer);
 tick = 0;
 install_int (timer, 1);

 gb_cur = 0;
 gbuf[0] = create_video_bitmap (MaxX, MaxY);
 gbuf[1] = create_video_bitmap (MaxX, MaxY);

 colour[COLOUR_BACK]     = 0x00000000;
 colour[COLOUR_BUILDER]  = 0x0060FF60;
 colour[COLOUR_MONSTER]  = 0x006060FF;
 colour[COLOUR_BORDER]   = 0x00404040;

 col_in [TYPE_BUILDER] = 0x0080D000;
 col_ex [TYPE_BUILDER] = 0x00408000;
 col_dir[TYPE_BUILDER] = 0x00F000F0;

 col_in [TYPE_MONSTER] = 0x008000D0;
 col_ex [TYPE_MONSTER] = 0x00400080;
 col_dir[TYPE_MONSTER] = 0x00F0F000;

 delta[TYPE_BUILDER] =  1;
 delta[TYPE_MONSTER] = -1;

 init_random ();
 strcpy (name, "anonymous");
}

void happy_end (void)
{
 remove_timer ();
 destroy_font (curfont);
 destroy_bitmap (gbuf[0]);
 destroy_bitmap (gbuf[1]);
 allegro_exit ();
 printf ("Successful termination.\n");
 fflush (stdout);
 exit (EXIT_SUCCESS);
}

bool move (void)
{
 int i, j;
 for (i = 0; i < MaxN; i++)
  if (crs[i] != NULL)
   crs[i] -> move ();

 for (i = 0; i < MaxN; i++)
  if (crs[i] != NULL)
   crs[i] -> check ();

 for (i = 0; i < MaxN; i++)
  if (crs[i] != NULL)
   crs[i] -> control ();

 for (j = 0; j < 4; j++)
  for (i = 0; i < MaxN; i++)
   if (crs[i] != NULL)
    crs[i] -> effect ();

 for (i = 0; i < MaxN; i++)
  if (crs[i] != NULL)
   if (crs[i] -> suffer ())
   {
    for (j = 0; j < 50; j++)
     crs[i] -> effect ();
    delete crs[i];
    crs[i] = NULL;
   }

 for (i = 1; i < MaxN; i++)
  if (crs[i] != NULL)
   return false;

 return true;
}

inline int boardcol (int boardval)
{
 int k, l;
 if (boardval > MidB)
 {
  k = boardval - MidB;
  l = MaxB - MidB;
  return (127 + (128 * k) / l) << 8;
 }
 else if (boardval < -MidB)
 {
  k = MidB + boardval;
  l = MaxB - MidB;
  return 128 + (128 * k) / l;
 }
 else
 {
  k = boardval + MidB;
  l = MidB << 1;
  return (((128 * k) / l) << 8) + (128 * (l - k)) / l;
 }
/*
 else if (boardval > 0)
 {
  k = MidB - boardval;
  l = MidB;
  return ((96 + (l - k)) << 8) + k * 3;
 }
 else
 {
  k = MidB + boardval;
  l = MidB;
  return ((k * 3) << 8) + 96 + (l - k);
 }
*/
}

void show_depth (void)
{
 int i;
 i = MidB;
 hline (gcur, 10, 250 - (i << 1), 40, 0x00D00000);
 hline (gcur, 10, 251 - (i << 1), 40, 0x00D00000);
 i = -MidB;
 hline (gcur, 10, 250 - (i << 1), 40, 0x00B08000);
 hline (gcur, 10, 251 - (i << 1), 40, 0x00B08000);
 for (i = -MaxB; i <= MaxB; i++)
 {
  hline (gcur, 15, 250 - (i << 1), 35, boardcol (i));
  hline (gcur, 15, 251 - (i << 1), 35, boardcol (i));
 }
 i = crs[0] -> z;
 hline (gcur, 12, 250 - (i << 1), 38, 0x00808080);
 hline (gcur, 12, 251 - (i << 1), 38, 0x00808080);
 i = MaxB - crs[0] -> v * (MaxB << 1);
 hline (gcur, 17, 250 - (i << 1), 24, 0x00F0F000);
 hline (gcur, 17, 251 - (i << 1), 24, 0x00F0F000);
 i = MaxB - crs[0] -> w * (MaxB << 1);
 hline (gcur, 26, 250 - (i << 1), 33, 0x0000F0F0);
 hline (gcur, 26, 251 - (i << 1), 33, 0x0000F0F0);
}

void show_score (void)
{
 textout_centre_ex (gcur, curfont, level_name, MaxX >> 1, 4,
  0x002020D0, 0x00B0B0B0);
 textprintf_centre_ex (gcur, curfont, MaxX >> 1, 26,
  0x00D0D000, 0x002020B0, "Time:  %10d", moves);
}

void draw (void)
{
 gb_cur ^= 1;
 gcur = gbuf[gb_cur];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BORDER]);

 int i, j;

 for (j = 0; j < BoardY; j++)
  for (i = 0; i < BoardX; i++)
   putpixel (gcur, ShiftX + i, ShiftY + j, boardcol (board[i][j]));

 for (i = 0; i < MaxN; i++)
  if (crs[i] != NULL)
   crs[i] -> draw ();
 show_depth ();
 show_score ();

 release_bitmap (gcur);
 show_video_bitmap (gcur);
}

void show_str (int x, int y, const char * msg)
{
 gb_cur ^= 1;
 gcur = gbuf[gb_cur];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BORDER]);
 textout_ex (gcur, curfont, msg, x, y, 0x00D0D000, 0x002020B0);
 release_bitmap (gcur);
 show_video_bitmap (gcur);
}

bool input_str (int x, int y, const char * prompt, char * str)
{
 int kb, kc, pos;
 char buf [MaxBuf];
 memset (str, 0, MaxInput);
 pos = 0;
 while (true)
 {
  sprintf (buf, "%s%-*s", prompt, MaxInput - 1, str);
  show_str (x, y, buf);
  rest (1);
  while (keypressed ())
  {
   kb = ureadkey (&kc);
   if (kc == KEY_ENTER)
   {
    return true;
   }
   else if (kc == KEY_ESC)
   {
    return false;
   }
   else if (kc == KEY_BACKSPACE)
   {
    if (pos > 0)
     str[--pos] = '\0';
   }
   else if (kb >= 32 && kb <= 127)
   {
    if (pos < MaxInput - 1)
    str[pos++] = kb;
   }
  }
 }
 return !true && !false;
}

void show_message (const char * msg)
{
 gb_cur ^= 1;
 gcur = gbuf[gb_cur];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BORDER]);
 textout_centre_ex (gcur, curfont, msg, MaxX >> 1, MaxY >> 1,
  0x00D0D000, 0x002020B0);
 release_bitmap (gcur);
 show_video_bitmap (gcur);
}

void show_level_start (const char * msg)
{
 gb_cur ^= 1;
 gcur = gbuf[gb_cur];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BORDER]);
 if (rndvalue (2))
 {
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 225,
   colour[COLOUR_BUILDER], colour[COLOUR_BORDER], "Builder:");
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 247,
   0xD0D020, colour[COLOUR_BORDER], "\"%s\"", str_startb[rndvalue (NumStr)]);
 }
 else
 {
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 225,
   colour[COLOUR_MONSTER], colour[COLOUR_BORDER], "Monster:");
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 247,
   0xD0D020, colour[COLOUR_BORDER], "\"%s\"", str_startm[rndvalue (NumStr)]);
 }
 textout_centre_ex (gcur, curfont, msg, MaxX >> 1, 275,
  0x00D0D000, 0x002020B0);
 release_bitmap (gcur);
 show_video_bitmap (gcur);
}

void init_ch_board (int builders)
{
 assert (builders > 0 && builders <= MaxN);
 int i, j;
 real sz;

 memset (crs, 0, sizeof (crs));
 sz = 10.0;
 i = 0;
 crs[i] = new creature (sz * 3.0 + rndvalue () * (BoardX - sz * 6.0),
  sz * 3.0 + rndvalue () * (BoardY - sz * 6.0),
  0.5, (rndvalue (ANum) * M2PI) / ANum, sz, TYPE_MONSTER, AI_NONE);

 for (i = 1; i <= builders; i++)
 {
  sz = 5.0;
  real xc, yc;
  do
  {
   xc = sz + rndvalue () * (BoardX - sz * 2.0);
   yc = sz + rndvalue () * (BoardY - sz * 2.0);
  }
  while (hypot (xc - crs[0] -> x, yc - crs[0] -> y) < 30.0);
  crs[i] = new creature (xc, yc, 1.0, (rndvalue (ANum) * M2PI) / ANum,
   sz, TYPE_BUILDER, rndvalue (AI_NUM - 2) + 2);
 }

 memset (board, 0, sizeof (board));
/*
 for (i = 0; i < BoardX; i++)
  for (j = 0; j < BoardY; j++)
   board[i][j] = rndvalue ((MaxB << 1) + 1) - MaxB;
 int k;
 for (k = 0; k < 25; k++)
  for (i = 0; i < BoardX; i++)
   for (j = 0; j < BoardY; j++)
    recur_effect_up (i, j),
    recur_effect_down (i, j);
*/

 for (j = 0; j < 150; j++)
  for (i = 1; i < MaxN; i++)
   if (crs[i] != NULL)
    crs[i] -> effect ();

 real x0 = crs[0] -> x, y0 = crs[0] -> y;
 for (i = 0; i < 10; i++)
 {
  for (j = 0; j < 350; j++)
   crs[0] -> effect ();
  crs[0] -> x = x0 + ((rndvalue (2) << 1) - 1) *
   (crs[0] -> s + rndvalue () * crs[0] -> s);
  crs[0] -> y = y0 + ((rndvalue (2) << 1) - 1) *
   (crs[0] -> s + rndvalue () * crs[0] -> s);
 }

 for (i = 0; i < MaxN; i++)
  if (crs[i] != NULL)
   crs[i] -> suffer ();
}


void change_name (void)
{
 if (input_str (10, MaxY >> 1, "Enter your name: ", buf))
 {
  strcpy (name, buf);
  if (!strcmp (name, ""))
   strcpy (name, "anonymous");
 }
}

void show_ch_score (void)
{
 int i, cf, cb;

 gb_cur ^= 1;
 gcur = gbuf[gb_cur];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BORDER]);

 textprintf_centre_ex (gcur, curfont, MaxX >> 1, 40,
  0x00D0D000, 0x002020B0, "High Scores for level:");
 textprintf_centre_ex (gcur, curfont, MaxX >> 1, 65,
  0x002020D0, 0x00B0B0B0, level_name);
 for (i = 0; i < MaxScore; i++)
  if (ch_score[level][i] < MaxC)
  {
   if (score_pos == i) cf = 0x00FF2020, cb = 0x004040D0;
                  else cf = 0x00D0D020, cb = 0x002020B0;
   textprintf_centre_ex (gcur, curfont, MaxX >> 1, 100 + i * 25,
    cf, cb, "%-*s %10d", MaxInput,
    ch_sname[level][i], ch_score[level][i]);
  }

 if (score_pos == -1)
 {
  i = 10;
  cf = 0x00FF2020, cb = 0x004040D0;
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 110 + i * 25,
   cf, cb, "%-*s %10d", MaxInput, name, moves);
 }
 if (win)
 {
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 400,
   colour[COLOUR_BUILDER], colour[COLOUR_BORDER], "Builder:");
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 422,
   0xD0D020, colour[COLOUR_BORDER], "\"%s\"", str_win   [rndvalue (NumStr)]);
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 450,
   0x00D0D000, 0x002020B0, "You win! Press ENTER to continue.");
 }
 else if (lose)
 {
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 400,
   colour[COLOUR_BUILDER], colour[COLOUR_BORDER], "Builder:");
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 422,
   0xD0D020, colour[COLOUR_BORDER], "\"%s\"", str_lose  [rndvalue (NumStr)]);
  textprintf_centre_ex (gcur, curfont, MaxX >> 1, 450,
   0x00D0D000, 0x002020B0, "You lose... Press ENTER to continue.");
 }

 release_bitmap (gcur);
 show_video_bitmap (gcur);
}

void challenge (void)
{
 int kb, kc;
 int player;
 player = 0;
 creature * & plr = crs[player];
 End1 = false;
 while (!End1)
 {
  init_ch_board (num_crs[level]);
  sprintf (level_name, "Challenge level %d", level + 1);
  sprintf (buf, "%s. Press ENTER to start...", level_name);
  show_level_start (buf);

  win = lose = false;
  while (true)
  {
   rest (1);
   while (!keypressed ()) ;
   if (key[KEY_ENTER]) break;
   if (key[KEY_ESC]) {End1 = true; break;}
   while (keypressed ()) kb = ureadkey (&kc);
  }

  while (!End1)
  {
   const int BATCH = 1, TICKS = 30;
   int i;
   for (i = 0; i < BATCH; i++)
   {
    if (move ())
    {
     win = true;
     delete crs[0];
     break;
    }
    if (plr == NULL)
    {
     lose = true;
     break;
    }
   }
   if (i < BATCH)
    break;
   draw ();
   while (tick < TICKS) rest (1);
   tick = 0;
   moves++;
   poll_kb ();
   if (key[KEY_UP]) plr -> accelerate ();
   if (key[KEY_DOWN]) plr -> brake ();
   if (key[KEY_LEFT]) plr -> turn_left ();
   if (key[KEY_RIGHT]) plr -> turn_right ();
   if (key[KEY_Q]) End1 = true;
   if (key[KEY_ESC]) End1 = true;
   while (keypressed ()) kb = ureadkey (&kc);
  }
  if (End1) break;

  if (win)
  {
   score_pos = insert_ch_score (level, moves);
   if (score_pos != -1)
    save_score ();
   show_ch_score ();
   level++;
  }
  else if (lose)
  {
   show_ch_score ();
   End1 = true;
  }
  while (true)
  {
   rest (1);
   while (!keypressed ()) ;
   if (key[KEY_ENTER]) break;
   if (key[KEY_ESC]) {End1 = true; break;}
   while (keypressed ()) kb = ureadkey (&kc);
  }
 }
}

void show_button (int x, int y, const char * msg)
{
 gb_cur ^= 1;
 gcur = gbuf[gb_cur];
 acquire_bitmap (gcur);
 clear_to_color (gcur, colour[COLOUR_BORDER]);
 textout_ex (gcur, curfont, msg, x, y, 0x00D0D000, 0x002020B0);
 release_bitmap (gcur);
 show_video_bitmap (gcur);
}

void show_menu (void)
{
}

void menu (void)
{
 change_name ();
 char buf2 [MaxBuf];
 int max_level;
 for (max_level = 0; max_level < MaxCh - 1 && ch_score[max_level][0] < MaxC;
  max_level++) ;
 level = 0;
 if (max_level > 0)
 {
  sprintf (buf2, "Starting level (%d-%d): ", 1, max_level + 1);
  if (input_str (10, MaxY >> 1, buf2, buf))
  {
   if (sscanf (buf, "%d", &level))
   {
    level--;
    if (level < 0 || level >= MaxCh)
     level = 0;
   }
   else level = 0;
  }
 }
 challenge ();
}

int main (void)
{
 init ();
 menu ();
 happy_end ();
 return 0;
}
END_OF_MAIN ()
